@@ -91,7 +91,7 @@ void catreturn()
 {
    /* Pop the object being edited from the push stack. */
 
-   popobject(NULL, NULL, NULL);
+   popobject(NULL, (pointertype)1, NULL);
 }
 
 /*------------------------------------------------------*/
@@ -328,7 +328,7 @@ void pagecatbutton(XButtonEvent *event)
 	 else if (event->button == Button1) {
 
 	    /* like catreturn(), but don't actually go to the popped page */
-	    objectdeselect();
+	    unselect_all();
 	    eventmode = NORMAL_MODE;
 	    if (mode == PAGELIB) {
 	       newpage(bpage);
@@ -341,7 +341,7 @@ void pagecatbutton(XButtonEvent *event)
 	 }
 	 else {  /* Button2 */
 	    if (mode == PAGELIB)    /* No such method for LIBLIB is defined. */
-	       objectselect(OBJECT);
+	       select_add_element(OBJINST);
 	 }
       }
    }
@@ -399,8 +399,8 @@ void pageinstpos(short mode, short tpage, objinstptr drawinst, int gxsize,
 objinstptr newpageinst(objectptr pageobj)
 {
    objinstptr newinst = (objinstptr) malloc(sizeof(objinst));
-   objectdefaults(newinst, pageobj, 0, 0);
-   newinst->type = OBJECT;
+   instancedefaults(newinst, pageobj, 0, 0);
+   newinst->type = OBJINST;
    newinst->color = DEFAULTCOLOR;
    return newinst;
 }
@@ -445,7 +445,7 @@ void composepagelib(short mode)
    /* from the list.							 */
 
    for (pgen = directory->plist; pgen < directory->plist + directory->parts; pgen++)
-      if ((*pgen)->type == OBJECT) *pgen = NULL;
+      if (IS_OBJINST(*pgen)) *pgen = NULL;
 
    reset(directory, NORMAL);
 
@@ -547,7 +547,7 @@ void updatepagelib(short mode, short tpage)
 
    for (i = 0; i < libinst->parts; i++) {
       gelem = libinst->plist + i;
-      if ((*gelem)->type == OBJECT) {
+      if (IS_OBJINST(*gelem)) {
 	 pinst = TOOBJINST(gelem);
          if (pinst->thisobject == compobj) {
 	    /* recalculate scale and position of the object instance */
@@ -630,7 +630,7 @@ void pagecatmove(XKeyEvent *event)
       }
    }
 
-   objectdeselect();
+   unselect_all();
    composelib(PAGELIB);
    drawarea(NULL, NULL, NULL);
 }
@@ -651,6 +651,9 @@ void composelib(short mode)
    short fval;
    short llx, lly, width, height;
 
+   int totalarea, targetwidth;
+   double scale;
+
    /* Also make composelib() a wrapper for composepagelib() */
    if ((mode > FONTLIB) && (mode < LIBRARY)) {
       composepagelib(mode);
@@ -662,7 +665,7 @@ void composelib(short mode)
    /* call reset(), we find the pointer to the instance and NULL it.	*/
    
    for (pgen = libpage->plist; pgen < libpage->plist + libpage->parts; pgen++)
-      if ((*pgen)->type == OBJECT) *pgen = NULL;
+      if (IS_OBJINST(*pgen)) *pgen = NULL;
 
    reset(libpage, NORMAL);
 
@@ -674,6 +677,35 @@ void composelib(short mode)
 
    fval = findhelvetica();
 
+   /* experimental:  attempt to produce a library with the same aspect  */
+   /* ratio as the drawing window.                                      */
+
+   totalarea = 0;
+   for (spec = xobjs.userlibs[mode - LIBRARY].instlist; spec != NULL;
+                spec = spec->next) {
+      libobj = spec->thisinst->thisobject;
+
+      /* "Hidden" objects are not drawn */
+      if (libobj->hidden == True) continue;
+
+      drawinst = spec->thisinst;
+      drawinst->position.x = 0;
+      drawinst->position.y = 0;
+
+      /* Get the bounding box of the instance in the page's coordinate system */
+      calcinstbbox((genericptr *)(&drawinst), &llx, &lly, &width, &height);
+      width -= llx;	/* convert urx to width */
+      height -= lly;	/* convert ury to height */
+      width += 30;	/* space padding */
+      height += 30;	/* height padding */
+      if (width < 200) width = 200;	/* minimum box width */
+      if (height < 220) height = 220;	/* minimum box height */
+      totalarea += (width * height);
+   }
+
+   scale = (double)totalarea / (double)(areastruct.width * areastruct.height);
+   targetwidth = (int)(sqrt(scale) * (double)areastruct.width);
+
    /* generate the list of object instances and their labels */
 
    for (spec = xobjs.userlibs[mode - LIBRARY].instlist; spec != NULL;
@@ -695,7 +727,8 @@ void composelib(short mode)
       /* Determine the area needed on the page to draw the object */
 
       nxpos = xpos + ((width > 170) ? width + 30 : 200);
-      if ((nxpos > (areastruct.width << 1)) && (xpos > 0)) {
+      /* if ((nxpos > (areastruct.width << 1)) && (xpos > 0)) { */
+      if ((nxpos > targetwidth) && (xpos > 0)) {
 	 nxpos -= xpos; 
 	 xpos = 0;
 	 ypos -= nypos;
@@ -771,7 +804,7 @@ short finddepend(objinstptr libobj, objectptr **compobjp)
 		     
          for (testobj = (*compobj)->plist; testobj < (*compobj)->plist
 	          + (*compobj)->parts; testobj++) {
-	    if ((*testobj)->type == OBJECT) {
+	    if (IS_OBJINST(*testobj)) {
 	       if (TOOBJINST(testobj)->thisobject == libobj->thisobject) return 2;
 	    }
 	 }
@@ -786,7 +819,7 @@ short finddepend(objinstptr libobj, objectptr **compobjp)
       *compobjp = compobj;
       for (testobj = (*compobj)->plist; testobj < (*compobj)->plist
 	       + (*compobj)->parts; testobj++) {
-	 if ((*testobj)->type == OBJECT) {
+	 if (IS_OBJINST(*testobj)) {
 	    if (TOOBJINST(testobj)->thisobject == libobj->thisobject) return 1;
 	 }
       }
@@ -1123,7 +1156,7 @@ void catmove(XKeyEvent *event)
 	 else if (areastruct.save.y <=
 		xobjs.libtop[i + LIBRARY]->thisobject->bbox.lowerleft.y +
 		xobjs.libtop[i + LIBRARY]->thisobject->bbox.height) {
-	    objectdeselect();	
+	    unselect_all();	
 	    Wprintf("Could not find appropriate place to insert object");
 	    return;
 	 }
@@ -1133,7 +1166,7 @@ void catmove(XKeyEvent *event)
       linkedlistinsertafter(&(xobjs.userlibs[i].instlist), s1, s2);
    }
 
-   objectdeselect();
+   unselect_all();
    if ((i = is_library(topobject)) >= 0) composelib(LIBRARY + i);
 
    drawarea(NULL, NULL, NULL);
@@ -1188,8 +1221,10 @@ void copycat()
       for (ops = oldobj->params; ops != NULL; ops = ops->next) {
 	 newops = (oparamptr)malloc(sizeof(oparam));
 	 newops->next = (*newobj)->params;
+	 newops->key = strdup(ops->key);
 	 (*newobj)->params = newops;
 	 newops->type = ops->type;
+	 newops->which = ops->which;
 	 switch (ops->type) {
 	    case XC_INT:
 	       newops->parameter.ivalue = ops->parameter.ivalue;
@@ -1221,7 +1256,7 @@ void copycat()
       (*newobj)->plist = (genericptr *)malloc(sizeof(genericptr));
 
       for (i = 0; i < oldobj->parts; i++) {
-	 switch((*(oldobj->plist + i))->type) {
+	 switch(ELEMENTTYPE(*(oldobj->plist + i))) {
 	    case(PATH): {
 	       register pathptr *npath, cpath = TOPATH(oldobj->plist + i);
 	       register genericptr *gpart;
@@ -1234,7 +1269,7 @@ void copycat()
                (*npath)->passed = NULL;
 	       for (gpart = cpath->plist; gpart < cpath->plist + cpath->parts;
 			gpart++) {
-		  switch((*gpart)->type){
+		  switch(ELEMENTTYPE(*gpart)){
 		     case ARC: {
 			arcptr copyarc = TOARC(gpart);
 			arcptr *newarc;
@@ -1287,41 +1322,15 @@ void copycat()
 
 	       clabel = TOLABEL(oldobj->plist + i);
 	       NEW_LABEL(nlabel, (*newobj));
-	       (*nlabel)->rotation = clabel->rotation;
-	       (*nlabel)->color = clabel->color;
-	       (*nlabel)->position.x = clabel->position.x;
-	       (*nlabel)->position.y = clabel->position.y;
-	       (*nlabel)->scale = clabel->scale;
-	       (*nlabel)->justify = clabel->justify;
-	       (*nlabel)->string = stringcopy(clabel->string);
-	       (*nlabel)->passed = NULL;
-	       (*nlabel)->pin = clabel->pin;
+	       labelcopy(*nlabel, clabel);
                } break;
       
-	    case(OBJECT): {
+	    case(OBJINST): {
 	       register objinstptr *ninst, cinst;
 
 	       cinst = TOOBJINST(oldobj->plist + i);
 	       NEW_OBJINST(ninst, (*newobj));
-	       (*ninst)->rotation = cinst->rotation;
-	       (*ninst)->color = cinst->color;
-	       (*ninst)->position.x = cinst->position.x;
-	       (*ninst)->position.y = cinst->position.y;
-	       (*ninst)->scale = cinst->scale;
-	       (*ninst)->thisobject = cinst->thisobject;
-	       (*ninst)->passed = NULL;
-
-	       /* Library instance's parameter list is always default (null) */
-	       (*ninst)->params = NULL;
-	       (*ninst)->bbox.lowerleft.x = cinst->thisobject->bbox.lowerleft.x;
-	       (*ninst)->bbox.lowerleft.y = cinst->thisobject->bbox.lowerleft.y;
-	       (*ninst)->bbox.width = cinst->thisobject->bbox.width;
-	       (*ninst)->bbox.height = cinst->thisobject->bbox.height;
-
-	       (*ninst)->schembbox = NULL;
-
-	       /* here---deal with attached schematics */
-
+	       instcopy(*ninst, cinst);
                } break;
          }
 	 (*newobj)->parts++;
@@ -1332,7 +1341,7 @@ void copycat()
    addtoinstlist(USERLIB - LIBRARY, *newobj, FALSE);
 
    composelib(USERLIB);
-   objectdeselect();
+   unselect_all();
 
    if (areastruct.topinstance == xobjs.libtop[USERLIB])
       drawarea(NULL, NULL, NULL);
@@ -1364,7 +1373,7 @@ void catbutton(u_int mode, XButtonEvent *event)
 	
       for (libobj = (objinstptr *)topobject->plist; libobj <
 		(objinstptr *)topobject->plist + topobject->parts; libobj++) {
-	 if ((*libobj)->type == OBJECT) {
+	 if (IS_OBJINST(*libobj)) {
 
  	    ocentx = (*libobj)->position.x + (*libobj)->bbox.lowerleft.x
 	        + ((*libobj)->bbox.width >> 1);
@@ -1468,11 +1477,13 @@ void catbutton(u_int mode, XButtonEvent *event)
                   Tk_CreateEventHandler(areastruct.area, PointerMotionMask,
 			(Tk_EventProc *)xctk_drag, NULL);
 #endif
+	 	  register_for_undo(XCF_Library_Pop, UNDO_MORE, areastruct.topinstance,
+				areastruct.selectlist, areastruct.selects);
                   catreturn();
                }
 
-               /* If button2 was pressed, select and stay in the catalog. */
-	       /* Could just do "objectselect" here, but would not cover   */
+               /* If button2 was pressed, select and stay in the catalog.  */
+	       /* Could just do "select_element" here, but would not cover */
 	       /* the entire area in the directory surrounding the object. */
 
 	       else {
